'use strict';

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _createForOfIteratorHelper(o, allowArrayLike) {
  var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"];

  if (!it) {
    if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
      if (it) o = it;
      var i = 0;

      var F = function () {};

      return {
        s: F,
        n: function () {
          if (i >= o.length) return {
            done: true
          };
          return {
            done: false,
            value: o[i++]
          };
        },
        e: function (e) {
          throw e;
        },
        f: F
      };
    }

    throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  var normalCompletion = true,
      didErr = false,
      err;
  return {
    s: function () {
      it = it.call(o);
    },
    n: function () {
      var step = it.next();
      normalCompletion = step.done;
      return step;
    },
    e: function (e) {
      didErr = true;
      err = e;
    },
    f: function () {
      try {
        if (!normalCompletion && it.return != null) it.return();
      } finally {
        if (didErr) throw err;
      }
    }
  };
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  Object.defineProperty(Constructor, "prototype", {
    writable: false
  });
  return Constructor;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function entityIsArchetype(entity, query) {
  return query.every(function (name) {
    return name in entity;
  });
}

var ListenerRegistry = /*#__PURE__*/function () {
  function ListenerRegistry() {
    _classCallCheck(this, ListenerRegistry);

    _defineProperty(this, "listeners", new Set());
  }

  _createClass(ListenerRegistry, [{
    key: "on",
    value: function on(listener) {
      this.listeners.add(listener);
    }
  }, {
    key: "off",
    value: function off(listener) {
      this.listeners["delete"](listener);
    }
  }, {
    key: "invoke",
    value: function invoke() {
      for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      this.listeners.forEach(function (l) {
        return l.apply(void 0, args);
      });
    }
  }]);

  return ListenerRegistry;
}();

var Archetype = /*#__PURE__*/function () {
  /** A list of entities belonging to this archetype. */

  /** Listeners on this event are invoked when an entity is added to this archetype's index. */

  /** Listeners on this event are invoked when an entity is removed from this archetype's index. */
  function Archetype(query) {
    _classCallCheck(this, Archetype);

    _defineProperty(this, "entities", new Array());

    _defineProperty(this, "onEntityAdded", new ListenerRegistry());

    _defineProperty(this, "onEntityRemoved", new ListenerRegistry());

    this.query = query;
  }

  _createClass(Archetype, [{
    key: "indexEntity",
    value: function indexEntity(entity) {
      var isArchetype = entityIsArchetype(entity, this.query);
      var pos = this.entities.indexOf(entity, 0);

      if (isArchetype && pos < 0) {
        this.entities.push(entity);
        this.onEntityAdded.invoke(entity);
      } else if (!isArchetype && pos >= 0) {
        this.entities.splice(pos, 1);
        this.onEntityRemoved.invoke(entity);
      }
    }
  }, {
    key: "removeEntity",
    value: function removeEntity(entity) {
      var pos = this.entities.indexOf(entity, 0);

      if (pos >= 0) {
        this.entities.splice(pos, 1);
        this.onEntityRemoved.invoke(entity);
      }
    }
  }]);

  return Archetype;
}();

function commandQueue() {
  var queue = new Array();

  function add(command) {
    queue.push(command);
  }

  function flush() {
    queue.forEach(function (fun) {
      return fun();
    });
    queue.length = 0;
  }

  function clear() {
    queue.length = 0;
  }

  return {
    add: add,
    flush: flush,
    clear: clear
  };
}

function idGenerator() {
  var startId = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0;
  var nextId = startId;
  return function () {
    return nextId++;
  };
}

/**
 * Normalize a list of component names, removing blank/undefined component names and sorting the resulting list.
 */
function normalizeComponentList(names) {
  return names === null || names === void 0 ? void 0 : names.filter(function (n) {
    return typeof n === "string" && n !== "";
  }).sort();
}

/**
 * Normalize an archetype by sorting the component names it references.
 */

function normalizeArchetype(query) {
  return normalizeComponentList(query);
}

/**
 * A base interface for entities, which are just normal JavaScript objects with
 * any number of properties, each of which represents a single component. When
 * using hmecs, it is recommended to create your own type representing your
 * game's entities, and pass that to `createWorld` for full typing support.
 */

/**
 * A tag is just an "empty" component. For convenience and nicer type support, we're
 * providing a Tag type and constant, both of which are simply equal to `true`.
 */
var Tag = true;
var World = /*#__PURE__*/function () {
  function World() {
    var _this = this;

    _classCallCheck(this, World);

    _defineProperty(this, "entities", new Array());

    _defineProperty(this, "nextId", idGenerator(1));

    _defineProperty(this, "archetypes", new Map());

    _defineProperty(this, "createEntity", function () {
      var entity = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

      /* If there already is an ID, raise an error. */
      if ("id" in entity) throw "Attempted to add an entity that aleady had an 'id' component.";
      /* Assign an ID */

      entity.id = _this.nextId();
      /* Store the entity... */

      _this.entities.push(entity);
      /* ...and add it to relevant indices. */


      _this.indexEntity(entity);

      return entity;
    });

    _defineProperty(this, "destroyEntity", function (entity) {
      var pos = _this.entities.indexOf(entity, 0);
      /* Sanity check */


      if (pos < 0) return;
      /* Remove it from our global list of entities */

      _this.entities.splice(pos, 1);
      /* Remove entity from all archetypes */


      var _iterator = _createForOfIteratorHelper(_this.archetypes.values()),
          _step;

      try {
        for (_iterator.s(); !(_step = _iterator.n()).done;) {
          var archetype = _step.value;
          archetype.removeEntity(entity);
        }
        /* Remove its id component */

      } catch (err) {
        _iterator.e(err);
      } finally {
        _iterator.f();
      }

      delete entity.id;
    });

    _defineProperty(this, "addComponent", function (entity, name, data) {
      if (name in entity) {
        throw "Tried to add component \"".concat(name, "\" to an entity that already has it.");
      }
      /* TODO: checking entity ownership like this is likely to slow us down quite a lot, so eventually we'll want something smarter here. */


      if (!_this.entities.includes(entity)) {
        throw "Tried to add component \"".concat(name, "\" to an entity that is not managed by this world.");
      }
      /* Add the component */


      entity[name] = data;
      /* Trigger a reindexing of the entity */

      _this.indexEntity(entity);
    });

    _defineProperty(this, "removeComponent", function (entity) {
      for (var _len = arguments.length, components = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        components[_key - 1] = arguments[_key];
      }

      /* TODO: checking entity ownership like this is likely to slow us down quite a lot, so eventually we'll want something smarter here. */
      if (!_this.entities.includes(entity)) {
        throw "Tried to remove ".concat(components, " from an entity that is not managed by this world.");
      }

      for (var _i = 0, _components = components; _i < _components.length; _i++) {
        var name = _components[_i];

        if (!(name in entity)) {
          throw "Tried to remove component \"".concat(name, " from an entity that doesn't have it.");
        }

        delete entity[name];
      }

      _this.indexEntity(entity);
    });

    _defineProperty(this, "queuedCommands", commandQueue());

    _defineProperty(this, "queue", {
      createEntity: function createEntity(entity) {
        _this.queuedCommands.add(function () {
          return _this.createEntity(entity);
        });

        return entity;
      },
      destroyEntity: function destroyEntity(entity) {
        _this.queuedCommands.add(function () {
          return _this.destroyEntity(entity);
        });
      },
      addComponent: function addComponent(entity, name, data) {
        _this.queuedCommands.add(function () {
          return _this.addComponent(entity, name, data);
        });
      },
      removeComponent: function removeComponent(entity) {
        for (var _len2 = arguments.length, names = new Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
          names[_key2 - 1] = arguments[_key2];
        }

        _this.queuedCommands.add(function () {
          return _this.removeComponent.apply(_this, [entity].concat(names));
        });
      },
      flush: function flush() {
        _this.queuedCommands.flush();
      }
    });
  }

  _createClass(World, [{
    key: "archetype",
    value: function archetype() {
      for (var _len3 = arguments.length, query = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
        query[_key3] = arguments[_key3];
      }

      var normalizedQuery = normalizeArchetype(query);
      /* We may already have an archetype representing the same query */

      var stringifiedQuery = JSON.stringify(normalizedQuery);
      if (this.archetypes.has(stringifiedQuery)) return this.archetypes.get(stringifiedQuery);
      /* Once we reach this point, we need to create a new archetype. */

      var archetype = new Archetype(normalizedQuery);
      this.archetypes.set(stringifiedQuery, archetype);

      var _iterator2 = _createForOfIteratorHelper(this.entities),
          _step2;

      try {
        for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
          var entity = _step2.value;
          archetype.indexEntity(entity);
        }
      } catch (err) {
        _iterator2.e(err);
      } finally {
        _iterator2.f();
      }

      return archetype;
    }
  }, {
    key: "indexEntity",
    value: function indexEntity(entity) {
      if (this.entities.includes(entity)) {
        var _iterator3 = _createForOfIteratorHelper(this.archetypes.values()),
            _step3;

        try {
          for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
            var archetype = _step3.value;
            archetype.indexEntity(entity);
          }
        } catch (err) {
          _iterator3.e(err);
        } finally {
          _iterator3.f();
        }
      }
    }
    /* MUTATION FUNCTIONS */

  }, {
    key: "clear",
    value:
    /**
     * Clear the entire world by discarding all entities and indices.
     */
    function clear() {
      /* Remove all entities */
      this.entities.length = 0;
      /* Remove all archetype indices */

      this.archetypes.clear();
      /* Clear queue */

      this.queuedCommands.clear();
    }
  }]);

  return World;
}();

exports.Archetype = Archetype;
exports.Tag = Tag;
exports.World = World;
exports._createForOfIteratorHelper = _createForOfIteratorHelper;
exports._unsupportedIterableToArray = _unsupportedIterableToArray;
