'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var World = require('../../dist/World-58a43038.cjs.dev.js');
var React = require('react');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var React__default = /*#__PURE__*/_interopDefault(React);

function useData(fn) {
  var ref = React.useRef();

  if (!ref.current) {
    ref.current = fn();
  }

  return ref.current;
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  var _i = arr == null ? null : typeof Symbol !== "undefined" && arr[Symbol.iterator] || arr["@@iterator"];

  if (_i == null) return;
  var _arr = [];
  var _n = true;
  var _d = false;

  var _s, _e;

  try {
    for (_i = _i.call(arr); !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || World._unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function useRerender() {
  var _useState = React.useState(0),
      _useState2 = _slicedToArray(_useState, 2);
      _useState2[0];
      var setVersion = _useState2[1];

  return React.useCallback(function () {
    setVersion(function (v) {
      return v + 1;
    });
  }, []);
}

function createECS() {
  var world = new World.World();
  var EntityContext = /*#__PURE__*/React.createContext(null);
  /**
   * A React component to either create a new entity, or represent an existing entity so
   * it can be enhanced with additional components (see the <Component> component.)
   */

  var Entity = function Entity(_ref) {
    var existingEntity = _ref.entity,
        children = _ref.children;

    /* Reuse the specified entity or create a new one */
    var entity = useData(function () {
      return existingEntity !== null && existingEntity !== void 0 ? existingEntity : world.createEntity();
    });
    /* If the entity was freshly created, manage its presence in the ECS world. */

    React.useEffect(function () {
      if (existingEntity) return;
      return function () {
        return world.destroyEntity(entity);
      };
    }, [entity]);
    /* Provide a context with the entity so <Component> components can be wired up. */

    return /*#__PURE__*/React__default["default"].createElement(EntityContext.Provider, {
      value: entity
    }, typeof children === "function" ? children(entity) : children);
  };

  var MemoizedEntity = /*#__PURE__*/React.memo(function (_ref2) {
    var entity = _ref2.entity,
        children = _ref2.children;
    return /*#__PURE__*/React__default["default"].createElement(Entity, {
      entity: entity,
      key: entity.id
    }, typeof children === "function" ? children(entity) : children);
  }, function (a, b) {
    return a.entity === b.entity;
  });

  var Entities = function Entities(_ref3) {
    var entities = _ref3.entities,
        _ref3$memoize = _ref3.memoize,
        memoize = _ref3$memoize === void 0 ? false : _ref3$memoize,
        children = _ref3.children;
    var Klass = memoize ? MemoizedEntity : Entity;
    return /*#__PURE__*/React__default["default"].createElement(React__default["default"].Fragment, null, entities.map(function (entity) {
      return /*#__PURE__*/React__default["default"].createElement(Klass, {
        entity: entity,
        key: entity.id,
        children: children
      });
    }));
  };

  var Collection = function Collection(_ref4) {
    var _ref4$initial = _ref4.initial,
        initial = _ref4$initial === void 0 ? 0 : _ref4$initial,
        _ref4$memoize = _ref4.memoize,
        memoize = _ref4$memoize === void 0 ? false : _ref4$memoize,
        tag = _ref4.tag,
        children = _ref4.children;

    var _useArchetype = useArchetype(tag),
        entities = _useArchetype.entities;

    React.useEffect(function () {
      /* When firing up, create the requested number of entities. */

      for (var i = 0; i < initial; i++) {
        var _entity = world.createEntity();

        world.addComponent(_entity, tag, World.Tag);
      }
      /* When shutting down, purge all the entities we have created. */


      return function () {
        var _iterator = World._createForOfIteratorHelper(entities),
            _step;

        try {
          for (_iterator.s(); !(_step = _iterator.n()).done;) {
            var _entity2 = _step.value;
            world.destroyEntity(_entity2);
          }
        } catch (err) {
          _iterator.e(err);
        } finally {
          _iterator.f();
        }
      };
    }, [tag, initial]);
    return /*#__PURE__*/React__default["default"].createElement(Entities, {
      entities: entities,
      memoize: memoize
    }, children);
  };
  /**
   * Return the current entity context.
   */


  function useEntity() {
    return React.useContext(EntityContext);
  }
  /**
   * Declaratively add a component to an entity.
   */


  function Component(props) {
    var entity = useEntity();
    var ref = React.useRef();
    /* Warn the user that passing multiple children is not allowed. */

    if ("children" in props && Array.isArray(props.children)) {
      throw new Error("<Component> will only accept a single React child.");
    }

    React.useEffect(function () {
      world.addComponent(entity, props.name, "data" in props ? props.data : ref.current);
      return function () {
        /* The entity might already have been destroyed, so let's check. */
        if ("id" in entity) {
          world.removeComponent(entity, props.name);
        }
      };
    }, [entity, props.name, "data" in props && props.data]);
    return /*#__PURE__*/React__default["default"].createElement(React__default["default"].Fragment, null, "children" in props && /*#__PURE__*/React.cloneElement(typeof props.children === "function" ? props.children(entity) : props.children, {
      ref: ref
    }));
  }
  /**
   * Return the entities of the specified archetype and subscribe this component
   * to it, making it re-render when entities are added to or removed from it.
   */


  function useArchetype() {
    for (var _len = arguments.length, query = new Array(_len), _key = 0; _key < _len; _key++) {
      query[_key] = arguments[_key];
    }

    var rerender = useRerender();
    var archetype = useData(function () {
      return world.archetype.apply(world, query);
    });
    React.useEffect(function () {
      archetype.onEntityAdded.on(rerender);
      archetype.onEntityRemoved.on(rerender);
      /* We need to rerender at least once, because other effects might have set up
         new entities before we had a chance to register our listeners. */

      rerender();
      return function () {
        archetype.onEntityAdded.off(rerender);
        archetype.onEntityRemoved.off(rerender);
      };
    }, [archetype]);
    return archetype;
  }

  return {
    world: world,
    useArchetype: useArchetype,
    useEntity: useEntity,
    Entity: Entity,
    Component: Component,
    MemoizedEntity: MemoizedEntity,
    Entities: Entities,
    Collection: Collection
  };
}

exports.createECS = createECS;
